#include "xmltk.h"
#include "SIndex.hpp"

class CStatHandler : public CTSAXContentHandler
{
public:
    CL_STDMETHOD(startDocument) ();
    CL_STDMETHOD(startElement) (XTOKEN xtName);
    CL_STDMETHOD(endElement) (XTOKEN xtName);

    CStatHandler()
    {
        CTSAXContentHandler();

        m_cElems         = 0;
        m_uiCurrentDepth = 0;
        m_uiMaxDepth     = 0;
	}

    unsigned int GetElementCount() { return m_cElems; }
    unsigned int GetMaxDepth() { return m_uiMaxDepth; }
    
private:
    unsigned int m_cElems;
    unsigned int m_uiCurrentDepth;
    unsigned int m_uiMaxDepth;
};

bool CStatHandler::startDocument()
{
    m_cElems         = 0;
    m_uiCurrentDepth = 0;
    m_uiMaxDepth     = 0;
    return true;
}

bool CStatHandler::startElement(XTOKEN xtName)
{
    m_cElems++;
    m_uiCurrentDepth++;
    if (m_uiCurrentDepth > m_uiMaxDepth)
        m_uiMaxDepth = m_uiCurrentDepth;
    return true;
}

bool CStatHandler::endElement(XTOKEN xtName)
{
    m_uiCurrentDepth--;
    return true;
}

class CCreateSIXHandler : public CTSAXContentHandler
{
public:
    CCreateSIXHandler(): CTSAXContentHandler()
    {
        m_pSIX = NULL;
        m_pXmlParser = NULL;
    }

    CL_STDMETHOD(startElement) (XTOKEN xtName);
    CL_STDMETHOD(endElement) (XTOKEN xtName);

    void SetSIX(SIndex * pSIX) { m_pSIX = pSIX; }
    void SetXmlParser(IParse2TSAX * pXmlParser) { m_pXmlParser = pXmlParser; }

private:
    SIndex * m_pSIX;
    IParse2TSAX * m_pXmlParser;
};

bool CCreateSIXHandler::startElement(XTOKEN xtName)
{
    m_pSIX->startElement(m_pXmlParser->getSrcOffset());
    return true;
}

bool CCreateSIXHandler::endElement(XTOKEN xtName)
{
    m_pSIX->endElement(m_pXmlParser->getSrcOffset());
    return true;
}

void Usage()
{
    cout << "Usage: createSindex [-t threshold] xmlfile" << endl;
}

int main(int argc, char* argv[])
{
    unsigned int threshold = 0;
    char * pszXmlFile = NULL;

    if (argc == 2)
    {
        if (!strcmp(argv[1], "-h"))
        {
            Usage();
            return 1;
        }
	    
        pszXmlFile = argv[1];
    }
    else if (argc == 4 && !strcmp(argv[1], "-t"))
    {
        // An index creation threshold was specified.
        threshold = atoi(argv[2]);
	    pszXmlFile = argv[3];
    }
    else
    {
        Usage();
        return 1;
    }

    InitGlobalTokenTable();
    try 
    {
        // 1st phase: Collect statistics of XML data.
        CStatHandler statHandler;
        {
            // Parse the xml file.
            CXmlToolkit xmltk(pszXmlFile, &statHandler);
            xmltk.Parse();
        }

    	// 2nd phase: Create the streaming index.
        CCreateSIXHandler sixHandler;
        {
            SIndex * pSIX = new SIndex(pszXmlFile, threshold);
            pSIX->setiTable(statHandler.GetElementCount(), 0);
            pSIX->setiPath(statHandler.GetMaxDepth(), 0);
	        pSIX->init();
            sixHandler.SetSIX(pSIX);

            CXmlToolkit xmltk(pszXmlFile, &sixHandler);
            sixHandler.SetXmlParser(xmltk.GetXmlParser());
            xmltk.Parse();

            pSIX->writeFile();
	        ATOMICDELETE(pSIX);
        }
    }
    catch (CXmlToolkitException &e)
    {
        printf("Error: %s", e.PszMessage());
    }
    CleanupGlobalTokenTable();
}

