#ifndef _XMLTK_H
#define _XMLTK_H

#ifdef WIN32
#pragma warning(disable:4786)
#endif

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <assert.h>
#include "stl.h"
#include "xmltk_util.h"
#include "xmltkobj.h"

#if !defined(__XPATH_H__) && !defined(__AUTOMATA_H__)
#include "../xpathDFA/Variable.h"
#endif  // __NODE_H__ && __AUTOMATA_H__

IParse2TSAX   * CreateXmlParser();
IXmlWriter    * CreateXmlWriter(int iIndent = 0);
IDfaFilter    * CreateDfaFilter(IFilteredTSAXHandler * pHandler = NULL);

IFileStream   * CreateFileStream();
IMemoryStream * CreateMemoryStream();

extern ITokenTable *g_ptt;

//
// NOTE: CreateXPathSMFactory takes only XPath expressions
// without qualifiers or XPath expressions with qualifiers on
// attributes only.  e.g., /bib/book or /bib/book[@id] but not
// /bib/book[/author].  For the latter expression, you must
// first parse it into two fragments, then create a factory
// for each fragment.
//
IClassFactoryCL *CreateXPathSMFactory(char *pszXPath);

BOOL IsEqualCLIID(const CLIID *riid1, const CLIID *riid2);

bool freadMultiByteUINT(IStream *pstm, unsigned int *puInt);
void fwriteMultiByteUINT(IStream *pstm, unsigned int uInt);

void funread(void *buffer, size_t size, size_t count, FILE *stream);

#ifdef DEBUG
void myassert(bool bCondition);
#else
#define myassert(b) {}
#endif  // DEBUG

int GetVariableDepth(Variable *pv);
//int QueryProcessMemoryUsage();

void ParseStream(IStream *pstm, ITSAXContentHandler *pch);

// TODO: We need to generalize this concept of output to a generic ITSAXContentHandler.
void WriteToStdOut(IStream * pstOut);

enum ErrorCodes
{
    ERROR_UNEXPECTED = 0,
    ERROR_NOT_IMPLEMENTED,
    ERROR_INPUT_FILE,
    ERROR_OUT_OF_MEMORY,
    ERROR_OUT_OF_BOUNDS,
    LAST_ERROR_CODE, // All the errors need to be defined before this.
};

class CXmlToolkitException
{
private:
    ErrorCodes   m_errorCode;
    char       * m_pszMessage;

    static const char * SZ_MESSAGES[LAST_ERROR_CODE]; // All the error messages, indexed by error code.

public:
    // CXmlToolkitException constructors.
    CXmlToolkitException(ErrorCodes errorCode); // Pre-defined error message.
    CXmlToolkitException(char * pszMessage);    // Custom error message.

    // Accessors.
    ErrorCodes ErrorCode() { return m_errorCode; }
    const char * PszMessage();
};

class CXmlToolkit
{
public:
    // CXmlToolkit constructors and destructor.
    CXmlToolkit(char * pszFile, ITSAXContentHandler * pHandler);
    virtual ~CXmlToolkit();

    // Perform the parsing operation.
    bool Parse();

    // Return the parser. Some programs require that.
    IParse2TSAX * GetXmlParser() { return m_pXmlParser; }

private:
    IFileStream          * m_pfsInput;   // File stream on the input.
    ITSAXContentHandler  * m_pHandler;   // This is the handler passed in by the user program.
    IParse2TSAX          * m_pXmlParser; // Actual XML parser.
};

#endif  // _XMLTK_H
