#include "xmltk.h"
#include "depth2offset.h"

Depth2offset::Depth2offset(const char * pszXmlFile)
{
    m_pPath          = NULL;
    m_uiMaxDepth     = 0;
    m_uiMaxRowSize   = 0;
    m_uiCurrentDepth = 0;
    m_uiSkipDepth    = 0;
    m_state          = S_CONSTRUCT;

    if (pszXmlFile == NULL) 
    {
        m_pSixFile = new FileManager("stdout.six");
	}
	else
    {
        // We are assuming the file ends in ".xml".
        char * pszSixFile = strdup(pszXmlFile);
        strcpy(pszSixFile + strlen(pszSixFile) - strlen(".xml"), ".six");
	  
        m_pSixFile = new FileManager(pszSixFile);
		free(pszSixFile);
    }
}

Depth2offset::~Depth2offset()
{
    ATOMICDELETE(m_pSixFile);
    ATOMICDELETE(m_pPath);
}

void Depth2offset::Init()
{
    if (m_pSixFile->open()) 
    {
        m_state = S_ACTIVE;
        struct indexHeader ih;
        m_pSixFile->read(&ih, sizeof(struct indexHeader));
        m_uiMaxDepth = ih.depth;
        m_uiMaxRowSize = ih.rowSize;
        m_pPath = (struct indexTable *) calloc(m_uiMaxDepth, sizeof(struct indexTable));
        m_pSixFile->read(&m_pPath[m_uiCurrentDepth], sizeof(struct indexTable));
    }
}

void Depth2offset::startElement(unsigned int elementID) 
{
	if (m_state != S_ACTIVE) 
        return;

    if (m_uiSkipDepth > 0)
    {
        m_uiSkipDepth++;
	    return;
	}
	
    if (m_pPath[m_uiCurrentDepth].startOffset == elementID)
    {
	    m_uiCurrentDepth++;
	    return;
	}
	
    // There is no entry.
	if (m_pPath[m_uiCurrentDepth].startOffset > elementID)
    {
	    m_uiSkipDepth++;
	    return;
	}

    while (m_pSixFile->read(&m_pPath[m_uiCurrentDepth], sizeof(struct indexTable)))
    {
        if (m_pPath[m_uiCurrentDepth].startOffset < elementID) 
            continue;
        else if (m_pPath[m_uiCurrentDepth].startOffset == elementID)
        {
            // There is an entry.
            m_uiCurrentDepth++;
		    return;
	    }
	    else 
        {
            // No entry for the element, but already read a following entry.
            m_uiSkipDepth++;
		    return;
	    }
	}
    
    m_state = S_END_OF_INDEX;
}

void Depth2offset::endElement(unsigned int elementID)
{
	if (m_state != S_ACTIVE)
        return;

	if (m_uiSkipDepth > 0)
    {
        m_uiSkipDepth--;
        return;
    }
	
    if (m_uiCurrentDepth == 0)
    {
        cerr << "Stack is empty." << endl;
        return;
	}
    
    // There should be an entry.
	m_uiCurrentDepth--;

    // m_pPath[m_uiCurrentDepth+1] can store a following entry for a coming element.
	m_pPath[m_uiCurrentDepth]= m_pPath[m_uiCurrentDepth+1];
}

unsigned int Depth2offset::failSkip(unsigned int depth)
{
	if (m_state != S_ACTIVE || depth == 0) 
        return 0;
	
    if (depth > m_uiCurrentDepth)
    {
        cerr << "The specified skip depth (" << depth << ") exceeds the number of stacked elements (" << m_uiCurrentDepth << ")." << endl;
        return 0;
	}

    m_uiCurrentDepth -= depth;
    
    // m_pPath[m_uiCurrentDepth + depth] can store a following entry for a coming element.
	m_pPath[m_uiCurrentDepth+1] = m_pPath[m_uiCurrentDepth+depth];
	return (m_pPath[m_uiCurrentDepth].endOffsetinOrg);
}

unsigned int Depth2offset::skip(unsigned int depth, unsigned int currentAddress)
{
	if (m_state != S_ACTIVE || depth == 0)
        return 0;

    if (depth > m_uiCurrentDepth)
    {
        cerr << "The specified skip depth (" << depth << ") exceeds the number of stacked elements (" << m_uiCurrentDepth << ")." << endl;
        return 0;
    }

    m_uiCurrentDepth -= depth;
	
    // m_pPath[m_uiCurrentDepth+depth] can store a following entry for a coming element.
    m_pPath[m_uiCurrentDepth+1] = m_pPath[m_uiCurrentDepth + depth];
    return (m_pPath[m_uiCurrentDepth].startOffset + m_pPath[m_uiCurrentDepth].endOffsetinOrg - currentAddress);
}

void Depth2offset::print(void)
{
    m_pSixFile->seek(0, FileManager::T_SEEK_SET);
    struct indexHeader ih;
	m_pSixFile->read(&ih, sizeof(struct indexHeader));
	fprintf(stdout, "rowSize=%d, depth=%d\n", ih.rowSize, ih.depth);
    
    struct indexTable it;
    for (unsigned int i = 0 ; i < m_uiMaxRowSize ; i++)
    {
        m_pSixFile->read(&it, sizeof(struct indexTable));
        fprintf(stdout, "startElement=%d, endElement=%d\n", it.startOffset, it.endOffsetinOrg);
	}
}
