#include "xmltk.h"
#include "FileStream.h"

const char * CXmlToolkitException::SZ_MESSAGES[LAST_ERROR_CODE] =
{
    "Unexpected error. Please contact your support personnel. :)",
    "Functionality not supported in this version of the toolkit.",
    "Could not open input file.",
    "Out of memory.",
    "Array index out of bounds."
};

CXmlToolkitException::CXmlToolkitException(ErrorCodes errorCode)
{
    m_errorCode = errorCode;
    // The message will be filled in from the error code only when accessed.
    m_pszMessage = NULL;
}

CXmlToolkitException::CXmlToolkitException(char * pszMessage)
{
    m_errorCode = LAST_ERROR_CODE;
    m_pszMessage = pszMessage;
}

const char * CXmlToolkitException::PszMessage()
{
    if (m_pszMessage == NULL)
    {
        assert(m_errorCode < LAST_ERROR_CODE);
        m_pszMessage = const_cast<char *>(SZ_MESSAGES[m_errorCode]);
    }
    
    return m_pszMessage;
}

CXmlToolkit::CXmlToolkit(char * pszFile, ITSAXContentHandler * pHandler)
{
    // Handle the input. Set the input to stdin if no file specified.
    m_pfsInput = new CFileStream();
    if (pszFile != NULL)
    {
        if (!m_pfsInput->OpenFile(pszFile, "r"))
            throw CXmlToolkitException(ERROR_INPUT_FILE);
    }
    else
    {
        m_pfsInput->SetFile(stdin);
    }

    ASSIGN(m_pHandler, pHandler);
    m_pXmlParser = CreateXmlParser();
}

CXmlToolkit::~CXmlToolkit()
{
    ATOMICRELEASE(m_pXmlParser);
    ATOMICRELEASE(m_pHandler);
    ATOMICRELEASE(m_pfsInput);
}

bool CXmlToolkit::Parse()
{
    return m_pXmlParser->Parse(m_pfsInput, m_pHandler);
}