#include "xmltk.h"

#ifndef WIN32
#include <sys/types.h>
#include <sys/sysinfo.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#else
#include <process.h>
#include <fstream>
#endif

BOOL IsEqualCLIID(const CLIID *riid1, const CLIID *riid2)
{
    return (memcmp(riid1, riid2, sizeof(*riid1)) == 0);
}


static const unsigned char c_chMask = 0x7F;     // 0x7F = 0111 1111
static const unsigned char c_chContinue = 0x80; // 0x80 = 1000 0000

void fwriteMultiByteUINT(IStream *pstm, unsigned int uInt)
{
    unsigned char rgch[5];                          // 5 = ceil(32/7)
    int i;
    
    for (i = 0; i < ARRAYSIZE(rgch); i++)
    {
        rgch[i] = (uInt & c_chMask) | c_chContinue;
        uInt >>= 7;
        if (uInt <= 0)
        {
            break;
        }
    }

    rgch[0] &= ~c_chContinue;

    for ( ; i >= 0; i--)
    {
        pstm->WriteChar(rgch[i]);
    }
}

bool freadMultiByteUINT(IStream *pstm, unsigned int *puInt)
{
    *puInt = 0;
    int i;

    //
    // use a loop to guard against overflow
    //
    for (i = 0; i < 5; i++)     // 5 = ceil(32/7)
    {
        int ch = pstm->ReadChar();
        if (ch == EOF)
        {
            return false;
        }
        else
        {
            *puInt <<= 7;
            *puInt |= (ch & c_chMask);
            if (!(ch & c_chContinue))
            {
                return true;
            }
        }
    }

    return false;
}

void funread(void *buffer, size_t size, size_t count, FILE *stream)
{
    for (int i = count - 1; i >= 0; i--)
    {
        for (int j = (size/sizeof(char)) - 1; j >= 0; j--)
        {
            void *bufferT = (void*)((size_t)buffer + (size*i));
            int ch = ((char*)bufferT)[j];
            ungetc(ch, stream);
        }
    }
}

#ifdef DEBUG
void myassert(bool bCondition)
{
#ifdef WIN32
    if (!bCondition)
    {
        __asm
        {
            int 3;
        }
    }
#else
    assert(bCondition);
#endif
}
#endif // DEBUG

XString * XString::Copy()
{
    XString * pxsCopy = new XString();
    pxsCopy->pszString = new char[cchString + 1]; // We will null-terminate the string.
    strncpy(pxsCopy->pszString, pszString, cchString);
    (pxsCopy->pszString)[cchString] = '\0';
    pxsCopy->cchString = cchString;

    return pxsCopy;
}

// WARNING: This is a 'less than' operation, not an equality check!
int XString::Compare(XString * pxs1, XString * pxs2)
{
    if (pxs1 == NULL || pxs1->pszString == NULL)
        return (pxs2 == NULL || pxs2->pszString == NULL ? 0 : 1);
    
    if (pxs2 == NULL || pxs2->pszString == NULL)
        return -1;
    
    size_t cch1 = pxs1->cchString;
    size_t cch2 = pxs2->cchString;
    assert(cch1 != 0 && cch2 != 0);
    
    int cmp = strncmp(pxs1->pszString, pxs2->pszString, min(cch1, cch2));
    if (cmp == 0 && cch1 != cch2)
    {
        cmp = (cch1 < cch2 ? 1 : -1);
    }
    return cmp;
}

int GetVariableDepth(Variable *pv)
{
    int iDepth = 0;
    while ((pv = pv->getParent()) != NULL)
    {
        ++iDepth;
    }
    return iDepth;
}

//
// magic bytetok calculation taken from the source 
// code for top
//
inline unsigned long bytetok(unsigned long l)
{
    return (l + 512) >> 10;
}

inline unsigned long bytetobyte(unsigned long l)
{
    return l + 512;
}

#ifndef WIN32
int _QueryProcFileMemUsage()
{
    int cbUsed = 0;

    char szBuf[4096];
    sprintf(szBuf, "/proc/%d/stat", getpid());
    szBuf[ARRAYSIZE(szBuf)-1] = 0;

    int fd = open(szBuf, O_RDONLY);
    if (fd != -1)
    {
        //
        // my psychic powers tell me that the memory usage stat
        // is the 23rd token
        //
        read(fd, szBuf, sizeof(szBuf)-1);

        szBuf[ARRAYSIZE(szBuf)-1] = 0;

        char *psz = szBuf;
        int i;
        for (i = 0; i < 22; i++)
        {
            psz = strchr(psz, ' ') + 1;
        }
                  
        cbUsed = bytetobyte(strtoul(psz, NULL, 10));

	close(fd);
    }
    
    return cbUsed;
}

//
// QueryProcessMemoryUsage is based on top's read_one_proc_stat
//
int QueryProcessMemoryUsage()
{
#ifdef USE_SBRK
    static unsigned long s_baseproc = _QueryProcFileMemUsage();
    static unsigned long s_baseheap = (unsigned long) sbrk(0);

    return (unsigned long) s_baseproc + (unsigned long) sbrk(0) - s_baseheap;
#else // USE_SBRK

    return _QueryProcFileMemUsage();

#endif // USE_SBRK
}
#endif // !WIN32

void WriteToStdOut(IStream * pstOut)
{
    while (true)
    {
        char szOutput[ONEK + 1];
        long cbRead = pstOut->Read(szOutput, ONEK);
        if (cbRead == 0)
            break;

        szOutput[cbRead++] = '\0';
        fwrite(szOutput, cbRead, 1, stdout);
    }
}

CStringBuffer::CStringBuffer()
{
    m_fStack     = true;
    m_pchCurrent = m_szBuffer;
    m_pszBuffer  = NULL;
    m_cchBuffer  = INIT_BUFFER_SIZE;
}

CStringBuffer::~CStringBuffer()
{
    if (m_pszBuffer != NULL)
        delete [] m_pszBuffer;
}

void CStringBuffer::Clear()
{
    if (!m_fStack)
    {
        m_fStack = true;
        m_cchBuffer = INIT_BUFFER_SIZE;
        
        assert(m_pszBuffer != NULL);
        delete [] m_pszBuffer;
        m_pszBuffer = NULL;
    }
    
    assert(m_pszBuffer == NULL);
    m_pchCurrent = m_szBuffer;
}

void CStringBuffer::Write(char * pszText, int cchText)
{
    int cchRemaining = m_cchBuffer - (m_pchCurrent - (m_fStack ? m_szBuffer : m_pszBuffer)); 
    if (cchRemaining < cchText)
    {
        // We don't have enough room, allocate more.
        m_cchBuffer = max(m_cchBuffer, cchText) * 2;
        char * pszNewBuffer = new char[m_cchBuffer + 1]; // Always leave room for the null terminator.
            
        int cchToCopy = 0;
        if (m_fStack)
        {
            // Switch to dynamic mode.
            m_pszBuffer = pszNewBuffer;

            // Copy the contents of the stack-based buffer to the dynamic one.
            cchToCopy = (m_pchCurrent - m_szBuffer);
            strncpy(m_pszBuffer, m_szBuffer, cchToCopy);

            m_fStack = false;
        }
        else
        {
            // We are already in dynamic mode, just make the dynamic buffer bigger.
            cchToCopy = (m_pchCurrent - m_pszBuffer);
            strncpy(pszNewBuffer, m_pszBuffer, cchToCopy);
            delete m_pszBuffer;
            m_pszBuffer = pszNewBuffer;
        }

        // In both cases, we are now in dynamic mode.
        m_pchCurrent = m_pszBuffer + cchToCopy;
    }

    // Now we ensured that we have enough room.
    strncpy(m_pchCurrent, pszText, cchText);
    m_pchCurrent += cchText;
}

const char * CStringBuffer::SzString()
{
    *m_pchCurrent = '\0';
    return (m_fStack ? m_szBuffer : m_pszBuffer);
}