#include "xmltk.h"
#include "saxclient.h"

void SAXClient::HandleStartElement(const char * pszName)
{
    if (m_pIndexMgr != NULL)
        m_pIndexMgr->startElement(m_pXmlParser->getSrcOffset());
    
    XTOKEN xt = g_ptt->XTOKENFromStr(const_cast<char *>(pszName), XST_ELEMENT);
    m_pchOut->startElement(xt);
}

void SAXClient::HandleAttribute(const char * pszName, const char * pszValue)
{
    XTOKEN xtAttribName = g_ptt->XTOKENFromStr(const_cast<char *>(pszName), XST_ATTRIBUTE);
    m_pchOut->attribute(xtAttribName, const_cast<char *>(pszValue), strlen(pszValue));
}

void SAXClient::HandleCharacters(const char * pszChars, int cchChars)
{
    m_pchOut->characters(const_cast<char *>(pszChars), cchChars);
}

void SAXClient::HandleEndElement(const char * pszName)
{
    if (m_pIndexMgr != NULL)
        m_pIndexMgr->endElement(m_pXmlParser->getSrcOffset());
    
    XTOKEN xt = g_ptt->XTOKENFromStr(const_cast<char *>(pszName), XST_ELEMENT);
    m_pchOut->endElement(xt);
}

void SAXClient::HandleComment(char *str,int len,char iscont)
{
    // Handles a piece of comment data
}

void SAXClient::HandlePI(char *str,int len,char iscont)
{
    // Handles a piece of processing instruction data
}

void SAXClient::HandleDOCTYPE(char *str,int len,char iscont)
{
    // Handles a piece of DOCTYPE data
}

void SAXClient::HandleStartCDATA(void)
{
    m_pchOut->startCDATA();
}

void SAXClient::HandleEndCDATA(void)
{
    m_pchOut->endCDATA();
}

SAXClient::SAXClient(ITSAXContentHandler * pchOut, IParse2TSAX * pXmlParser, Depth2offset * pIndexMgr)
{
    m_pXmlParser  = pXmlParser;
    m_pIndexMgr   = pIndexMgr;
    ASSIGN(m_pchOut, pchOut);
}

SAXClient::~SAXClient()
{
    ATOMICRELEASE(m_pchOut);
}

#include "xmlparser.h"
#include "xml2tsax.h"

bool CXML2TSAX::QueryInterface(RCLIID riid, void **ppvObj)
{
    if (IsEqualCLIID(riid, &IID_IUnknownCL) ||
        IsEqualCLIID(riid, &IID_IParse2TSAX))
    {
        *ppvObj = static_cast<IParse2TSAX*>(this);
    }
    else
    {
        return false;
    }
    AddRef();
    return true;
}

ULONG CXML2TSAX::AddRef()
{
    return ++m_cRef;
}

ULONG CXML2TSAX::Release()
{
    --m_cRef;
    if (m_cRef > 0)
    {
        return m_cRef;
    }
    delete this;
    return 0;
}

#define CHECK_RET(x) if (!(fReturn = (x))) goto Cleanup;

bool CXML2TSAX::Parse(IStream *pstm, ITSAXContentHandler *pch)
{
    bool fReturn = false;

    IFileStream * pfs = NULL;
    CHECK_RET(pstm->QueryInterface(&IID_IFileStream, (void **) &pfs));

    try 
    {         
        CHECK_RET(m_xmlParser.OpenFile(pfs->GetFileName()) == 1);

        indexMgr = new Depth2offset(pfs->GetFileName());
	    indexMgr->Init();
        SAXClient saxclient(pch, this, indexMgr);
        
        pch->startDocument();
	    m_xmlParser.Parse(&saxclient);
        pch->endDocument();
    }
    catch (XMillException const *)
    {
        fprintf(stderr, "XML parser error. Is \"%s\" correct?\n", pfs->GetFileName());
    }

Cleanup:
    m_xmlParser.CloseFile();

    if (pfs != NULL)
        pfs->Release();

    return fReturn;
}
    
unsigned int CXML2TSAX::getSrcOffset(void)
{
    return m_xmlParser.GetSrcOffset();
}

bool CXML2TSAX::skip(unsigned int skipDepth)
{
  if (skipDepth == 0) return true;
  if (indexMgr && indexMgr->isAppliable()==true){
	unsigned int offset = indexMgr->skip(skipDepth, getSrcOffset());
	m_xmlParser.SkipReader(offset);
	return true;
  }
  else return false;
}

IParse2TSAX * CreateXmlParser()
{
    return new CXML2TSAX();
}