#include "xmltk.h"
#include "xmlparser.h"

XMLParser::XMLParser()
{
    Input();

    m_fSkip = false;

    // Create the Expat parser object.
    m_pExpatParser = XML_ParserCreate(NULL);

    // Set the handlers.
    XML_SetElementHandler(m_pExpatParser, StartElement, EndElement);
    XML_SetCharacterDataHandler(m_pExpatParser, Characters);
    XML_SetCdataSectionHandler(m_pExpatParser, StartCData, EndCData);
}

XMLParser::~XMLParser()
{
    if (m_pExpatParser != NULL)
        XML_ParserFree(m_pExpatParser);
}

void XMLParser::Parse(SAXClient * pClient)
{
    // Set the client as the user data argument to the callback functions of the parser.
    XML_SetUserData(m_pExpatParser, pClient);

    unsigned int cbRead = 0;
    do
    {
        char * pszBuffer = (char *) XML_GetBuffer(m_pExpatParser, FILEBUF_SIZE);
        if (pszBuffer == NULL)
            throw CXmlToolkitException(ERROR_OUT_OF_MEMORY);

        cbRead = ReadData(pszBuffer, FILEBUF_SIZE);
        
        if (XML_ParseBuffer(m_pExpatParser, cbRead, cbRead != FILEBUF_SIZE) == XML_STATUS_ERROR)
        {
            XML_Error error = XML_GetErrorCode(m_pExpatParser);
            throw CXmlToolkitException(const_cast<char *>(XML_ErrorString(error)));
        }
    } while (cbRead == FILEBUF_SIZE);
}

void XMLParser::SkipReader(unsigned int cbSkip)
{
    skipReaderSub(cbSkip, 0);
	m_fSkip = true;
}

void XMLCALL XMLParser::StartElement(void * pUserData, const XML_Char *name, const XML_Char ** ppAttribs)
{
    SAXClient * pSAXClient = (SAXClient *) pUserData;
    
    pSAXClient->HandleStartElement(name);
    if (ppAttribs != NULL)
    {
        while (*ppAttribs != NULL)
        {
            assert(*(ppAttribs + 1) != NULL);
            pSAXClient->HandleAttribute(*ppAttribs, *(ppAttribs + 1));
        }
    }
}

void XMLCALL XMLParser::EndElement(void * pUserData, const XML_Char *name)
{
    ((SAXClient *) pUserData)->HandleEndElement(name);
}

void XMLCALL XMLParser::Characters(void * pUserData, const XML_Char * pszChars, int cchChars)
{
    ((SAXClient *) pUserData)->HandleCharacters(pszChars, cchChars);
}

void XMLCALL XMLParser::StartCData(void * pUserData)
{
    ((SAXClient *) pUserData)->HandleStartCDATA();
}

void XMLCALL XMLParser::EndCData(void * pUserData)
{
    ((SAXClient *) pUserData)->HandleEndCDATA();
}
