#include "xmltk.h"

Variable *g_pvRoot = NULL;
Variable *g_pvElement = NULL;
bool g_bRecursive = false;
IDfaFilter *g_pfilter = NULL;
char *g_pszFileName = NULL;

class CXFlattenHandler : public CTSAXContentHandler
{
public:
	CL_STDMETHOD(startDocument) ();
	CL_STDMETHOD(endDocument) ();
	CL_STDMETHOD(startElement) (XTOKEN xtName);
	CL_STDMETHOD(endElement) (XTOKEN xtName);
	CL_STDMETHOD(attribute) (XTOKEN xtName, char *pszChars, int cchChars);
	CL_STDMETHOD(characters) (char *pszChars, int cchChars);
	CL_STDMETHOD(startCDATA) (void);
    CL_STDMETHOD(endCDATA) (void);

	CL_STDMETHOD(startContext) (Variable *pv);
	CL_STDMETHOD(endContext) (Variable *pv);

	CXFlattenHandler(ITSAXContentHandler *pchOut)
	{
	    m_bOpenTagSkipped = false;
	    m_bCloseTagAvailable = false;
	    m_bSkipOpenTag = m_bSkipCloseTag = false;
	    ASSIGN(m_pchOut, pchOut);

	    m_ContextDepth = 0;
	}

	virtual ~CXFlattenHandler()
	{
	    ATOMICRELEASE(m_pchOut);
	}

private:
	void _OnOpenElement(Variable *pv)
	{
	    m_bSkipOpenTag = true;
	}

	void _OnCloseElement()
	{
	    m_bSkipCloseTag = true;
	}

	bool m_bSkipOpenTag;
	bool m_bSkipCloseTag;
	bool m_bOpenTagSkipped;
	bool m_bCloseTagAvailable;

	int m_ContextDepth;

	ITSAXContentHandler *m_pchOut;

	XTOKEN m_xtoken;
};

bool CXFlattenHandler::startDocument()
{
    m_pchOut->startDocument();
    return true;
}

bool CXFlattenHandler::endDocument()
{
    if (m_bCloseTagAvailable)
    {
        m_bCloseTagAvailable = false;
        if (!m_bSkipCloseTag)
        {
            m_pchOut->endElement(m_xtoken);
    	}
        else
        {
            m_bSkipCloseTag = false;
        }
    }
    m_pchOut->endDocument();
    return true;
}

bool CXFlattenHandler::startElement(XTOKEN xtName)
{
    if (m_bCloseTagAvailable)
    {
        m_bCloseTagAvailable = false;
        if (!m_bSkipCloseTag)
        {
            m_pchOut->endElement(m_xtoken);
	    }
	    else
	    {
	        // skipping the close tag
	        //
	        m_bSkipCloseTag = false;
	    }
    }

    if (!m_bSkipOpenTag || m_bOpenTagSkipped)
    {
	    m_pchOut->startElement(xtName); 
    }

    if (m_bSkipOpenTag)
    {
        if (m_bOpenTagSkipped)
        {
            m_bSkipOpenTag = false;
            m_bOpenTagSkipped = false;
        }
        else
        {
            m_bOpenTagSkipped = true;
        }
    }

    return true;
}

bool CXFlattenHandler::endElement(XTOKEN xtName)
{
    if (m_bCloseTagAvailable)
    {
        m_bCloseTagAvailable = false;
        if (!m_bSkipCloseTag)
        {
            m_pchOut->endElement(m_xtoken);
        }
        else
        {
            // skipping the close tag
            //
            m_bSkipCloseTag = false;
        }
    }
    
    m_xtoken = xtName;
    m_bCloseTagAvailable = true;
    return true;
}

bool CXFlattenHandler::attribute(XTOKEN xtName, char *pszChars, int cchChars)
{
    if (!m_bSkipOpenTag)
    {
        m_pchOut->attribute(xtName, pszChars, cchChars);
    }
    return true;
}

bool CXFlattenHandler::characters(char *pszChars, int cchChars)
{
    m_pchOut->characters(pszChars, cchChars);
    return true;
}

bool CXFlattenHandler::startCDATA()
{
    m_pchOut->startCDATA();
    return true;
}

bool CXFlattenHandler::endCDATA()
{
    m_pchOut->endCDATA();
    return true;
}

bool CXFlattenHandler::startContext(Variable *pv)
{
    switch (GetVariableDepth(pv))
    {
	case 0:
        if (pv != g_pvRoot)
        {
            if (g_bRecursive || m_ContextDepth == 0)
            {
                _OnOpenElement(pv);
            }
		    m_ContextDepth++;
	    }
	    break;
	default:
	    break;
    }

    return true;
}

bool CXFlattenHandler::endContext(Variable *pv)
{
    switch (GetVariableDepth(pv))
    {
	case 0:
	    m_ContextDepth--;
	    if (g_bRecursive || m_ContextDepth == 0)
	    {
            _OnCloseElement();
	    }
	    break;
	default:
	    break;
    }
    
    return true;
}

bool FProcessInput(int argc, char * argv[])
{
    for (int iArg = 1 ; iArg < argc ; iArg++)
    {
        char * pszArg = argv[iArg];
        if (*pszArg != '-')
        {
            g_pszFileName = strdup(pszArg);
            return true;
        }
        pszArg++;

        if (pszArg[1] != '\0')
            return false;

        switch (*pszArg)
        {
        case 'r':
            g_bRecursive = true;
            break;
        case 'e':
            g_pvElement = g_pfilter->RegisterQuery(g_pvRoot, argv[++iArg], true);
            break;
        default:
            return false;
        }
    }

    if (g_pvElement == NULL)
        return false;

    return true;
}

int Cleanup()
{
    ATOMICFREE(g_pszFileName);

    ATOMICRELEASE(g_pfilter);
    CleanupGlobalTokenTable();

    return 0;
}

int main(int argc, char* argv[])
{
    InitGlobalTokenTable();
    g_pfilter = CreateDfaFilter();
    
    if (!FProcessInput(argc, argv))
    {
        cout << "Usage: xflatten [-r] -e <element to remove> [input file]\n";
        cout << "       -r: Recursive remove.\n";
        return Cleanup();
    }

    try
    {
        g_pvRoot = g_pfilter->RegisterQuery(NULL, "/", true);
        
        IXmlWriter * pchOut = CreateXmlWriter();
        pchOut->SetOutputFile(stdout);
        
        CXFlattenHandler handler(pchOut);
        g_pfilter->SetHandler(&handler);
        
        // Parse the xml file.
        CXmlToolkit xmltk(g_pszFileName, g_pfilter);
        g_pfilter->SetXmlParser(xmltk.GetXmlParser());
        xmltk.Parse();
	}
    catch (CXmlToolkitException &e)
    {
        printf("\n\n !! Error: %s !!", e.PszMessage());
    }

	return Cleanup();
}