#include "xmltk.h"
#include <string.h>

#define DEFAULT_GROUP_TAG "group"
#define DEFAULT_KEY_TAG "key"

#define ELEMENT_PRECEDENCE 1.0
#define KEY_PRECEDENCE 2.0

Variable *g_pvRoot;
char *g_pszGroupTag = NULL;
char *g_pszKeyTag = NULL;
XTOKEN g_xtGroupTag;
XTOKEN g_xtKeyTag;
Variable *g_pvActiveGroup = NULL;	// only one group can be active at a time

IXmlWriter * g_pxwKey = NULL;
string g_strKey = "";

class CElementAccumulator
{
public:
    CElementAccumulator()
    {
        m_pfsElement1 = CreateFileStream();

        m_pxwElement = CreateXmlWriter();
        m_pxwElement->SetOutputStream(m_pfsElement1);
    }

    virtual ITSAXContentHandler *GetHandlerForNewElement(ITSAXContentHandler *pchOut) = 0;

    virtual void OnCloseElement(ITSAXContentHandler *pchOut) = 0;

    void OutputOpenGroupTag(ITSAXContentHandler *pchOut)
    {
	pchOut->startElement(g_xtGroupTag);
    }

    void FlushElement(ITSAXContentHandler *pchOut)
    {
	    pfsCurElement()->Seek(0, SEEK_SET);
        WriteToStdOut(pfsCurElement());
	    pfsCurElement()->CloseFile();
    }

    virtual void CloseGroup(ITSAXContentHandler *pchOut) = 0;

    void OutputCloseGroupTag(ITSAXContentHandler *pchOut)
    {
	pchOut->endElement(g_xtGroupTag);
    }

    void PrepareElementFileStream()
    {
        if (pfsCurElement()->GetFile() == NULL)
        {
            pfsCurElement()->SetFile(tmpfile());
        }
        m_pxwElement->SetOutputStream(pfsCurElement());
    }

    IFileStream *pfsCurElement() { return m_pfsElement1; }
    
    virtual ~CElementAccumulator()
    {
	m_pfsElement1->CloseFile();

	ATOMICRELEASE(m_pfsElement1);
	ATOMICRELEASE(m_pxwElement);
    }

protected:
    IXmlWriter * m_pxwElement;
    IFileStream *m_pfsElement1;
};


class AccumulatorWithCount : public CElementAccumulator
{
public:
    AccumulatorWithCount(int groupSize)
    {
	m_groupSize = groupSize;
	m_elementsCollected = 0;
    }

    virtual ITSAXContentHandler *GetHandlerForNewElement(ITSAXContentHandler *pchOut)
    {
	if (m_elementsCollected == 0)
	{
	    OutputOpenGroupTag(pchOut);
	}
	
	PrepareElementFileStream();
	m_pxwElement->AddRef();
	return m_pxwElement;
    }

    virtual void OnCloseElement(ITSAXContentHandler *pchOut)
    {
	FlushElement(pchOut);
	m_elementsCollected++;
	if (m_elementsCollected == m_groupSize)
	{
	    CloseGroup(pchOut);
	    g_pvActiveGroup = NULL;
	}
    }

    virtual void CloseGroup(ITSAXContentHandler *pchOut)
    {
	m_elementsCollected = 0;
	OutputCloseGroupTag(pchOut);
    }

    virtual ~AccumulatorWithCount()
    {
    }

private:
    int m_groupSize;
    int m_elementsCollected;
};

class AccumulatorWithKey : public CElementAccumulator
{
public:
    AccumulatorWithKey()
    {
        m_pfsKey = CreateFileStream();
	
        m_pxwKey = CreateXmlWriter();
        m_pxwKey->SetOutputStream(m_pfsKey);

        m_strKey = NULL;
        m_bKeyFlushed = false;
    }

    virtual ITSAXContentHandler *GetHandlerForNewElement(ITSAXContentHandler *pchOut)
    {
        PrepareElementFileStream();
        m_pxwElement->AddRef();
        return m_pxwElement;
    }

    virtual void OnCloseElement(ITSAXContentHandler *pchOut)
    {
	if (m_strKey == NULL)
	{
	    OutputOpenGroupTag(pchOut);
	    FlushKey(pchOut);
	}
	else if (! CompareKeys())
	{
	    CloseGroup(pchOut);
	    OutputOpenGroupTag(pchOut);
	    FlushKey(pchOut);
	}

	FlushElement(pchOut);
	m_pfsKey->CloseFile();
    }

    virtual void CloseGroup(ITSAXContentHandler *pchOut)
    {
	if (m_strKey != NULL)
	{
	    delete m_strKey; m_strKey = NULL;
	}
	OutputCloseGroupTag(pchOut);
    }

    ITSAXContentHandler *GetHandlerForNewKey()
    {
	    PrepareKeyFileStream();
	    m_pxwKey->AddRef();
	    return m_pxwKey;
    }

    void PrepareKeyFileStream()
    {
	    if (m_pfsKey->GetFile() == NULL)
	    {
	        m_pfsKey->SetFile(tmpfile());
	    }

	    m_pxwKey->SetOutputStream(m_pfsKey);
    }

    void FlushKey(ITSAXContentHandler *pchOut)
    {
	// output the open tag <key>
	//
	pchOut->startElement(g_xtKeyTag);

	// output the key
	//
	if (m_pfsKey->GetFile())
	{
	    m_pfsKey->Seek(0, SEEK_SET);
        WriteToStdOut(m_pfsKey);
	}

	// output the close tag </key>
	// 
	pchOut->endElement(g_xtKeyTag);

	m_strKey = new string(g_strKey);
    }

    virtual ~AccumulatorWithKey()
    {
        m_pfsKey->CloseFile();
        ATOMICRELEASE(m_pfsKey);
        ATOMICRELEASE(m_pxwKey);
        ATOMICDELETE(m_strKey);
    }

    bool CompareKeys()
    {
        return (strcmp(g_strKey.c_str(), m_strKey->c_str()) == 0);
    }

private:
    IFileStream *m_pfsKey;
    IXmlWriter * m_pxwKey;

    string *m_strKey;

    bool m_bKeyFlushed;
};

class CXNestHandler : public IFilteredTSAXHandler
{
public:
    // *** IUnknownCL methods ***
    CL_STDMETHOD(QueryInterface) (RCLIID riid, void **ppvObj);
    CL_STDMETHOD_(ULONG,AddRef) ();
    CL_STDMETHOD_(ULONG,Release) ();
    
    // *** ITSAXContentHandler methods ***
    CL_STDMETHOD(startDocument) ();
    CL_STDMETHOD(endDocument) ();
    CL_STDMETHOD(startElement) (XTOKEN xtName);
    CL_STDMETHOD(endElement) (XTOKEN xtName);
    CL_STDMETHOD(attribute) (XTOKEN xtName, char *pszChars, int cchChars);
    CL_STDMETHOD(characters) (char *pszChars, int cchChars);
    CL_STDMETHOD(startCDATA) ();
    CL_STDMETHOD(endCDATA) ();

    // *** IFilteredTSAXHandler methods ***
    CL_STDMETHOD(startContext) (Variable *pv);
    CL_STDMETHOD(endContext) (Variable *pv);

    CXNestHandler(ITSAXContentHandler *pchOut)
    {
	m_pvElement = NULL;
	m_pchOut = pchOut;
        m_pchOut->AddRef();
	m_pchCur = pchOut;
	m_pchCur->AddRef();
	m_cRef = 1;
	m_bKeyContext = false;
    }

    virtual ~CXNestHandler()
    {
        ATOMICRELEASE(m_pchOut);
        ATOMICRELEASE(m_pchCur);
    }

private:
        
    void _OnOpenElement(Variable *pv)
    {
	if ((g_pvActiveGroup != NULL) && (g_pvActiveGroup != pv))
	{
	    CElementAccumulator *pec = (CElementAccumulator *)g_pvActiveGroup->getlParam();
	    pec->CloseGroup(m_pchOut);
	}
	g_pvActiveGroup = pv;
	m_pvElement = pv;
	CElementAccumulator *pec = (CElementAccumulator *)pv->getlParam();
	ATOMICRELEASE(m_pchCur);
	m_pchCur = pec->GetHandlerForNewElement(m_pchOut);

	// reset the key before we go into this element
	//
	g_strKey = "";
    }

    void _OnCloseElement(Variable *pv)
    {
	m_pvElement = NULL;
	CElementAccumulator *pec = (CElementAccumulator *)pv->getlParam();
	pec->OnCloseElement(m_pchOut);

    ATOMICRELEASE(m_pchCur);
    ASSIGN(m_pchCur, m_pchOut);
    }

    void _OnOpenKey(Variable *pv)
    {
	m_bKeyContext = true;
	AccumulatorWithKey *peck = (AccumulatorWithKey *)pv->getlParam();
	g_pxwKey = (IXmlWriter *) peck->GetHandlerForNewKey();
    }

    void _OnCloseKey(Variable *pv)
    {
	m_bKeyContext = false;
	ATOMICRELEASE(g_pxwKey);
    }
	
    Variable *m_pvElement;
    
    ITSAXContentHandler *m_pchOut;
    ITSAXContentHandler *m_pchCur;
    ULONG m_cRef;

    bool m_bKeyContext;
};

// *** IUnknown methods ***
bool CXNestHandler::QueryInterface(RCLIID riid, void **ppvObj)
{
    if (IsEqualCLIID(riid, &IID_IUnknownCL) ||
        IsEqualCLIID(riid, &IID_ITSAXContentHandler) ||
        IsEqualCLIID(riid, &IID_IFilteredTSAXHandler))
    {
        *ppvObj = static_cast<IFilteredTSAXHandler*>(this);
    }
    else
    {
        *ppvObj = NULL;
        return false;
    }
    
    AddRef();
    return true;
}

ULONG CXNestHandler::AddRef()
{
    return ++m_cRef;
}

ULONG CXNestHandler::Release()
{
    if (--m_cRef > 0)
    {
        return m_cRef;
    }
    delete this;
    return 0;
}

// *** ITSAXContentHandler methods ***
bool CXNestHandler::startDocument()
{
    m_pchCur->startDocument();
    return true;
}

bool CXNestHandler::endDocument()
{
    m_pchCur->endDocument();
    return true;
}

bool CXNestHandler::startElement(XTOKEN xtName)
{
    if ((g_pvActiveGroup != NULL) && (m_pvElement == NULL))
    {
	CElementAccumulator *pec = (CElementAccumulator *)g_pvActiveGroup->getlParam();
	pec->CloseGroup(m_pchOut);
	g_pvActiveGroup = NULL;
    }
#if defined(XMLTK_OUTPUT)	
    cerr << "startElement(" << g_ptt->StrFromXTOKEN(xtName) << ")" << endl;
#else
    m_pchCur->startElement(xtName); 
#endif

    if (m_bKeyContext)
        g_pxwKey->startElement(xtName);
    
    return true;
}

bool CXNestHandler::endElement(XTOKEN xtName)
{
    if ((g_pvActiveGroup != NULL) && (m_pvElement == NULL))
    {
	CElementAccumulator *pec = (CElementAccumulator *)g_pvActiveGroup->getlParam();
	pec->CloseGroup(m_pchOut);
	g_pvActiveGroup = NULL;
    }
#if defined(XMLTK_OUTPUT)	
    cerr << "endElement(" << g_ptt->StrFromXTOKEN(xtName) << ")" << endl;
#else
    m_pchCur->endElement(xtName);
#endif

    if (m_bKeyContext)
        g_pxwKey->endElement(xtName);
    
    return true;
}

bool CXNestHandler::attribute(XTOKEN xtName, char *pszChars, int cchChars)
{
#if defined(XMLTK_OUTPUT)	
    cerr << "attribute(" << g_ptt->StrFromXTOKEN(xtName) << ")" << endl;
#else
    m_pchCur->attribute(xtName, pszChars, cchChars);
#endif
    
    if (m_bKeyContext)
    {
        g_strKey += pszChars;
        g_pxwKey->attribute(xtName, pszChars, cchChars);
    }

    return true;
}

bool CXNestHandler::characters(char *pszChars, int cchChars)
{
#if defined(XMLTK_OUTPUT)
    cerr << "characters()" << endl;
#else
    m_pchCur->characters(pszChars, cchChars);
#endif
    
    if (m_bKeyContext)
    {
        g_strKey += pszChars;
        g_pxwKey->characters(pszChars, cchChars);
    }
    
    return true;
}

bool CXNestHandler::startCDATA()
{
    m_pchCur->startCDATA();

    if (m_bKeyContext)
        g_pxwKey->startCDATA();
    
    return true;
}

bool CXNestHandler::endCDATA()
{
    m_pchCur->endCDATA();

    if (m_bKeyContext)
        g_pxwKey->endCDATA();
    
    return true;
}

// *** IFilteredTSAXHandler methods ***
bool CXNestHandler::startContext(Variable *pv)
{
#if defined(XMLTK_OUTPUT)	
    cerr << "startContext(" << pv << ")" << endl;
#endif
    if (pv->getPrecedence() == ELEMENT_PRECEDENCE)
    {
	// check if there is an element for which neither count nor key was
	// specified
	if (pv->getlParam() == ((size_t) NULL))
	{
	    pv->setlParam((size_t) new AccumulatorWithCount(INT_MAX));
	}
	_OnOpenElement(pv);
    }
    else if (pv->getPrecedence() == KEY_PRECEDENCE)
    {
	_OnOpenKey(pv);
    }

    return true;
}
bool CXNestHandler::endContext(Variable *pv)
{
#if defined(XMLTK_OUTPUT)	
    cerr << "endContext(" << pv << ")" << endl;
#endif
    if (pv->getPrecedence() == ELEMENT_PRECEDENCE)
    {
	_OnCloseElement(pv);
    }
    else if (pv->getPrecedence() == KEY_PRECEDENCE)
    {
	_OnCloseKey(pv);
    }
	
    return true;
}

Variable *g_pvElement = NULL;
Variable *g_pvKey = NULL;
IDfaFilter *g_pfilter = NULL;
char *g_pszFileName = NULL;

void _UseGroupTag(char *psz)
{
    g_pszGroupTag = strdup(psz);
    g_xtGroupTag = g_ptt->XTOKENFromStr(g_pszGroupTag, XST_ELEMENT);
}

void _UseKeyTag(char *psz)
{
    g_pszKeyTag = strdup(psz);
    g_xtKeyTag = g_ptt->XTOKENFromStr(g_pszKeyTag, XST_ELEMENT);
}

void _UseElement(char *psz)
{
    g_pvElement = g_pfilter->RegisterQuery(g_pvRoot, psz, true);
    g_pvElement->setPrecedence(ELEMENT_PRECEDENCE);
    g_pvElement->setlParam((size_t) NULL);
}

void _UseCount(int i)
{
    if (i <= 0)
    {
	fprintf(stderr, "usage error - number specified with an element should be greater than zero\n");
	exit(0);
    }

    g_pvElement->setlParam((size_t) new AccumulatorWithCount(i));
}

void _UseKey(char *psz)
{
    g_pvKey = g_pfilter->RegisterQuery(g_pvElement, psz, true);
    g_pvKey->setPrecedence(KEY_PRECEDENCE);
    CElementAccumulator *pec;
    if ((pec = (AccumulatorWithKey *) g_pvElement->getlParam()) == NULL)
    {
	pec = new AccumulatorWithKey();
	g_pvElement->setlParam((size_t)pec);
    }
    g_pvKey->setlParam((size_t)pec);
}

bool FProcessInput(int argc, char * argv[])
{
    for (int iArg = 1 ; iArg < argc ; iArg++)
    {
        char * pszArg = argv[iArg];
        if (*pszArg != '-')
        {
            g_pszFileName = strdup(pszArg);
            return true;
        }
        pszArg++;
        
        switch (*pszArg)
        {
        case 'g':
            if (*++pszArg != 't')
				return false;
			_UseGroupTag(argv[++iArg]);
            break;
        case 'n':
            _UseCount(atoi(argv[++iArg]));
            break;
        case 'e':
            _UseElement(argv[++iArg]);
            break;
        case 'k':
            {
			char chNext = *++pszArg;
			if (chNext == 't')
			    _UseKeyTag(argv[++iArg]);
			else if (chNext == '\0')
			    _UseKey(argv[++iArg]);
			else
				return false;
            break;
            }
        case '?':
        default:
            return false;
        }
    }

    return true;
}

int Cleanup()
{
    ATOMICFREE(g_pszFileName);
	ATOMICFREE(g_pszKeyTag);
    ATOMICFREE(g_pszGroupTag);

    ATOMICRELEASE(g_pfilter);
    CleanupGlobalTokenTable();

    return 0;
}

int main(int argc, char* argv[])
{
    InitGlobalTokenTable();
    g_pfilter = CreateDfaFilter();

    if (!FProcessInput(argc, argv))
    {
        cout << "Usage: xnest [-gt <group-tag-name>] [-kt <key-tag-name>] -e <element to include>"
             << " (-k <key> | -n number) [input file]\n";
        return Cleanup();
    }
    
    try
    {
        g_pvRoot = g_pfilter->RegisterQuery(NULL, "/", true);
	    
        if (g_pszGroupTag == NULL)
            _UseGroupTag(DEFAULT_GROUP_TAG);
        if (g_pszKeyTag == NULL)
	        _UseKeyTag(DEFAULT_KEY_TAG);
	    
	    IXmlWriter * pchOut = CreateXmlWriter();
        pchOut->SetOutputFile(stdout);
        
        CXNestHandler handler(pchOut);
        g_pfilter->SetHandler(&handler);
        
        // Parse the xml file.
        CXmlToolkit xmltk(g_pszFileName, g_pfilter);
        g_pfilter->SetXmlParser(xmltk.GetXmlParser());
        xmltk.Parse();
	}
    catch (CXmlToolkitException &e)
    {
        printf("\n\n !! Error: %s !!", e.PszMessage());
    }
        
    return Cleanup();
}