#include "xmltk.h"

#define ELEMENT_PRECEDENCE 1.0
#define GROUP_PRECEDENCE 2.0

Variable *g_pvElement = NULL;
Variable *g_pvGroup = NULL;
Variable *g_pvRoot = NULL;
bool g_fPairTag = false;
XTOKEN g_xtPairTag;

class CXPairHandler : public CTSAXContentHandler
{
public:
    CL_STDMETHOD(startDocument) ();
    CL_STDMETHOD(endDocument) ();
    CL_STDMETHOD(startElement) (XTOKEN xtName);
    CL_STDMETHOD(endElement) (XTOKEN xtName);
    CL_STDMETHOD(attribute) (XTOKEN xtName, char *pszChars, int cchChars);
    CL_STDMETHOD(characters) (char *pszChars, int cchChars);
    CL_STDMETHOD(startCDATA) ();
    CL_STDMETHOD(endCDATA) ();

    CL_STDMETHOD(startContext) (Variable *pv);
    CL_STDMETHOD(endContext) (Variable *pv);

    CXPairHandler(IXmlWriter * pxwOut)
    {
        m_pvElement = NULL;
	    ASSIGN(m_pxwOut, pxwOut);
    }

    virtual ~CXPairHandler()
    {
        ATOMICRELEASE(m_pxwOut);
    }

private:

    void _OnOpenGroup(Variable *pv)
    {
        // The element to duplicate cannot be a child of the element to duplicate.
        if (m_pvElement != NULL)
            throw CXmlToolkitException("The element to duplicate cannot be a child of the element to duplicate.");

        m_pxwOut->startElement(g_xtPairTag);
	    m_pxwOut->FlushBuffer();
    }

    void _OnOpenElement(Variable *pv)
    {
        // The element to duplicate cannot contain itself.
        if (m_pvElement != NULL)
            throw CXmlToolkitException("The element to duplicate cannot contain itself.");

        m_pvElement = pv;
        // Set the output to buffering mode, with write through.
        m_pxwOut->SetBufferMode(true);
    }

    void _OnCloseElement(Variable *pv)
    {
        m_pvElement = NULL;
        m_pxwOut->ClearBufferMode();
    }

    Variable   * m_pvElement;
    IXmlWriter * m_pxwOut;
};


// *** ITSAXContentHandler methods ***
bool CXPairHandler::startDocument()
{
    m_pxwOut->startDocument();
    return true;
}

bool CXPairHandler::endDocument()
{
    m_pxwOut->endDocument();
    return true;
}

bool CXPairHandler::startElement(XTOKEN xtName)
{
    m_pxwOut->startElement(xtName);
    return true;
}

bool CXPairHandler::endElement(XTOKEN xtName)
{
    m_pxwOut->endElement(xtName);
    return true;
}

bool CXPairHandler::attribute(XTOKEN xtName, char *pszChars, int cchChars)
{
    m_pxwOut->attribute(xtName, pszChars, cchChars);
    return true;
}

bool CXPairHandler::characters(char *pszChars, int cchChars)
{
    m_pxwOut->characters(pszChars, cchChars);
    return true;
}

bool CXPairHandler::startCDATA()
{
    m_pxwOut->startCDATA();
    return true;
}

bool CXPairHandler::endCDATA()
{
    m_pxwOut->endCDATA();
    return true;
}

bool CXPairHandler::startContext(Variable *pv)
{
    if (pv->getPrecedence() == ELEMENT_PRECEDENCE)
    {
        _OnOpenElement(pv);
    }
    else if (pv->getPrecedence() == GROUP_PRECEDENCE)
    {
        _OnOpenGroup(pv);
    }
    return true;
}

bool CXPairHandler::endContext(Variable *pv)
{
    if (pv->getPrecedence() == ELEMENT_PRECEDENCE)
    {
        _OnCloseElement(pv);
    }
    else if (pv->getPrecedence() == GROUP_PRECEDENCE)
    {
        m_pxwOut->endElement(g_xtPairTag);
    }
    return true;
}

IDfaFilter *g_pfilter = NULL;
char *g_pszFileName = NULL;

void _UseGroup(char *psz)
{
    g_pvGroup = g_pfilter->RegisterQuery(g_pvRoot, psz, true);
    g_pvGroup->setPrecedence(GROUP_PRECEDENCE);
}

void _UseElement(char *psz)
{
    g_pvElement = g_pfilter->RegisterQuery(g_pvRoot, psz, true);
    g_pvElement->setPrecedence(ELEMENT_PRECEDENCE);
}

bool FProcessInput(int argc, char * argv[])
{
    bool fUserPairTag = false;

    for (int iArg = 1 ; iArg < argc ; iArg++)
    {
        char * pszArg = argv[iArg];
        if (*pszArg != '-')
        {
            g_pszFileName = strdup(pszArg);
            return true;
        }
        pszArg++;

        if (pszArg[1] != '\0')
            return false;

        switch (*pszArg)
        {
        case 'g':
            if (g_pvElement == NULL)
                return false;
			_UseGroup(argv[++iArg]);
            break;
        case 'e':
            _UseElement(argv[++iArg]);
            break;
        case 'p':
            fUserPairTag = true;
            g_xtPairTag = g_ptt->XTOKENFromStr(argv[++iArg], XST_ELEMENT);
            break;
        default:
            return false;
        }
    }

    if (g_pvGroup == NULL || g_pvElement == NULL)
        return false;

    if (!fUserPairTag)
        g_xtPairTag = g_ptt->XTOKENFromStr("pair", XST_ELEMENT);

    return true;
}

int Cleanup()
{
    ATOMICFREE(g_pszFileName);

    ATOMICRELEASE(g_pfilter);
    CleanupGlobalTokenTable();

    return 0;
}

int main(int argc, char* argv[])
{
    InitGlobalTokenTable();
    g_pfilter = CreateDfaFilter();
    
    if (!FProcessInput(argc, argv))
    {
        cout << "Usage: xpair [-p <pair tag>] -e <element to duplicate> -g <element to pair up> [input file]\n";
        return Cleanup();
    }

    try
    {
        g_pvRoot = g_pfilter->RegisterQuery(NULL, "/", true);
        
        IXmlWriter * pchOut = CreateXmlWriter();
        pchOut->SetOutputFile(stdout);
        
        CXPairHandler handler(pchOut);
        g_pfilter->SetHandler(&handler);
        
        // Parse the xml file.
        CXmlToolkit xmltk(g_pszFileName, g_pfilter);
        g_pfilter->SetXmlParser(xmltk.GetXmlParser());
        xmltk.Parse();
	}
    catch (CXmlToolkitException &e)
    {
        printf("\n\n !! Error: %s !!", e.PszMessage());
    }

	return Cleanup();
}
