/************************************************************************
// -*- mode: c++ -*-
//  copyright (C) 2001 Makoto Onizuka, University of Washington
//  $Id: List.cxx,v 1.1.1.1 2002/05/04 12:53:54 tjgreen Exp $
*************************************************************************/
#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define LIST_EMBODY
#include "List.h"

template<class T> 
List<T>::List()
{
    count = 0;
    next  = 0;
    size  = DEF_CHILD_COUNT;
    
    CHECK_ALLOC(list = (T **) malloc(size * sizeof(T*)));
    memset(list, 0, size * sizeof(T*));
}

template<class T> 
List<T>::List(unsigned int c): Base()
{
    count = 0;
    next  = 0;
    size  = c;
    
    CHECK_ALLOC(list = (T **) malloc(size * sizeof(T*)));
    memset(list, 0, size * sizeof(T*));
}

template<class T> 
List<T>::List(const List<T> * const target): Base(){
  size = target->size;
  count = next = target->count;
  CHECK_ALLOC(list = (T **) malloc(size * sizeof(T*)));
  memcpy(list, target->list, sizeof(T *) * size);
}

template<class T> 
List<T>::~List(){
//  pthread_mutex_destroy(&lock);
  free(list);
}

template<class T> 
const unsigned long List<T>::newItem()
{
    if (size == count)
    {
        // In case the list is full, extend the child list size.
        T ** oldList = list;
        CHECK_ALLOC(list = (T **) malloc(sizeof(T*) * (size+DEF_EXTEND_COUNT)));
      memcpy(list, oldList, sizeof(T *) * count);
      size += DEF_EXTEND_COUNT;
      for (unsigned int i = count; i < size; i++) list[i] = NULL;
      free (oldList);
  }
  if (list[next] == NULL) return next++;
  for (unsigned long i = 0; i < size; i++){
      if (list[i] == NULL){
	  next = i + 1;
	  return i;
      }
  }

  // We should never come here.
  throw CXmlToolkitException(ERROR_UNEXPECTED);
}

template<class T> 
T * List<T>::deleteItem(const unsigned long pos){
  T * ret = setItem(pos, NULL);
  for (unsigned int i = pos; i < count + 1; i++){
    list[i] = list[i+1];
  }
  return ret;
}

template<class T> 
T * List<T>::getItem(const unsigned long pos){
  if(pos >= size) return NULL;
  return list[pos];
}

template<class T> 
T * List<T>::setItem(const unsigned long pos, T * t)
{
    assert(pos < size);
  T * old = list[pos];
  if (list[pos] == NULL) count++; // vfȂƂɓꍇ
  if (t == NULL){		// 폜(NULL ݒ肷)ꍇ
    count--;
    if (pos==0) next = 0;
				// list[pos  - 1]  NULL łȂ
				// 폜ꏊ next ɂĂ
    else if ((pos > 0)&&(list[pos - 1] != NULL)) next = pos;
  }
  list[pos] = t;
  return old;
}

template<class T> 
void List<T>::delItems(void){
  for (unsigned long i = 0; i < size; i++){
      if (list[i] != NULL) {
	delete list[i];
	list[i] = 0;
      }
  }
  count = next = 0;
}
