%{
//	This is a lex program for xmatch processor.
//      copyright (C) 2001 Makoto Onizuka, University of Washington
//      $Id: xmatch.l,v 1.4 2002/09/09 06:45:56 monizuka Exp $
//

#include <string.h>
#include <stdlib.h>
#include "xmatch.tab.h"
#include "operator.h"
#ifdef WIN32
#include "xmatch.h"
#endif
extern int my_DFAinput(char *, int);
extern int DFAerror(char * mes);

#undef YY_INPUT
#define YY_INPUT(buf, result, max_size)(result = my_DFAinput(buf, max_size))

int lineCount(char * text, const char * del){
  int num = 0;
  char * top;
  char * target = top = strdup(text);
  while ((target = strstr(target, del))!=NULL){
    target += strlen(del);
    num++;
  }
  free(top);
  return num;
}

%}

%option noyywrap

%e 1200
%a 12000
%o 25000

Digit	        [0-9]
NameChar	[0-9a-zA-Z.\-_:]
TopNameChar	[a-zA-Z_:]

%%

 /* for INTEGER and FLOAT value */
<INITIAL>[0-9]+          {DFAlval.intval = atoi(DFAtext);
                         return INTEGER;}
<INITIAL>{Digit}*"."{Digit}+ {DFAlval.fltval = atof(DFAtext);
                         return FLOAT;}

 /* for comparison operator */
<INITIAL>"="               {DFAlval.subtok = MATH_OPR_EQ;  return OPERATOR;}
<INITIAL>"!="              {DFAlval.subtok = MATH_OPR_NEQ; return OPERATOR;}
<INITIAL>"<"               {DFAlval.subtok = MATH_OPR_LES; return OPERATOR;}
<INITIAL>">"               {DFAlval.subtok = MATH_OPR_GRE; return OPERATOR;}
<INITIAL>"<="              {DFAlval.subtok = MATH_OPR_LEQ; return OPERATOR;}
<INITIAL>">="              {DFAlval.subtok = MATH_OPR_GEQ; return OPERATOR;}
<INITIAL>"starts-with("	   {DFAlval.subtok = FUNC_STARTS;  return FUNCTION;}
<INITIAL>"contains("       {DFAlval.subtok = FUNC_CONTAINS;return FUNCTION;}

 /* for mathematical calculation */
<INITIAL>"div"             {return DIV;}
<INITIAL>"mod"             {return MOD;}

 /* for logical AND OR */
<INITIAL>[aA][nN][dD]      {return AND;}
<INITIAL>[oO][rR]          {return OR;}

 /* functions */
<INITIAL>"text("	   {return TEXT;}

 /* for XPath expression (including mathematical operation) */
<INITIAL>"//"              {return DSLASH;}
<INITIAL>"/"               {return SLASH;}
<INITIAL>[\[\].,()|*+\-@]   {return DFAtext[0];}

<INITIAL>{TopNameChar}{NameChar}*   {DFAlval.strval = DFAtext;
                         DFAlineno += lineCount(DFAtext, "\n");
                         return EREF;} /* element simple path */
<INITIAL>@{TopNameChar}{NameChar}*   {DFAlval.strval = DFAtext;
                         DFAlineno += lineCount(DFAtext, "\n");
                         return AREF;} /* attribute simple path */
 /* lex expression for a registerd function call should come here. */

 /* literal expression */
<INITIAL>'[^'\n]*'          {
#ifndef WIN32
     int c = input();
#else
     int c = DFAinput();
#endif WIN32
     unput(c);
     if (c != '\''){
         DFAlval.strval = DFAtext;
         DFAlval.strval[strlen(DFAlval.strval) - 1] = '\0';
         DFAlval.strval += 1;
         return STRING;
     }
     else
         yymore();
}

<INITIAL>\n		{DFAlineno++;} /* whitespace */
<INITIAL>[ \t\r]+	        ;	    /* whitespace */
<INITIAL>.		{if (strlen(DFAtext) != 0)
			   return DFAerror("lex: An invalid character in XPATH expression.");
		        }
%%
