// BUGS:
// * If an error occurs, the interfaces don't get cleaned up.
//   Solution: We should implement auto-pointers.

#include "xmltk.h"

class CXRunHandler : public CTSAXContentHandler
{
public:
    CL_STDMETHOD(startElement) (XTOKEN xtName);
    CL_STDMETHOD(endElement) (XTOKEN xtName);
    CL_STDMETHOD(attribute) (XTOKEN xtName, char *pszChars, int cchChars);
    CL_STDMETHOD(characters) (char *pszChars, int cchChars);
    CL_STDMETHOD(startCDATA) ();
    CL_STDMETHOD(endCDATA) ();    

    CXRunHandler(ITSAXContentHandler *pchOut) {
	  m_pchOut = pchOut;
	  m_pchOut->AddRef();
	}
    virtual ~CXRunHandler(){
	  ATOMICRELEASE(m_pchOut); 
	}

    ITSAXContentHandler *m_pchOut;
};

bool CXRunHandler::startElement(XTOKEN xtName)
{
    m_pchOut->startElement(xtName); 
    return true;
}

bool CXRunHandler::endElement(XTOKEN xtName)
{
    m_pchOut->endElement(xtName);
    return true;
}

bool CXRunHandler::attribute(XTOKEN xtName, char *pszChars, int cchChars)
{
    m_pchOut->attribute(xtName, pszChars, cchChars);
    return true;
}

bool CXRunHandler::characters(char *pszChars, int cchChars)
{
    m_pchOut->characters(pszChars, cchChars);
    return true;
}

bool CXRunHandler::startCDATA()
{
    m_pchOut->startCDATA();
    return true;
}

bool CXRunHandler::endCDATA()
{
    m_pchOut->endCDATA();
    return true;
}

int main(int argc, char* argv[])
{
    if (argc != 3)
    {
        cout << "Usage: " << argv[0] << " xpathExpr xmlfile" << endl;
        return 1;
    }

    InitGlobalTokenTable();
    try 
    {
        // Create my handler object that works on a stdout stream.
        IXmlWriter * pchOut = CreateXmlWriter(4);
        pchOut->SetOutputFile(stdout);
        CXRunHandler handler(pchOut);

        // Create a new XPath processor, setting the handler on it.
        IDfaFilter * pdfa = CreateDfaFilter(&handler);
        // Register the query with the processor. RegisterQuery will throw if something goes wrong.
        pdfa->RegisterQuery(NULL, argv[1], true);

        // Parse the xml file.
        CXmlToolkit xmltk(argv[2], pdfa);
        pdfa->SetXmlParser(xmltk.GetXmlParser());
        xmltk.Parse();

        // Cleanup.
        ATOMICRELEASE(pdfa);
        ATOMICRELEASE(pchOut);
    }
    catch (CXmlToolkitException &e)
    {
        printf("Error: %s", e.PszMessage());
    }
    CleanupGlobalTokenTable();
}